"""Constants for the library."""
from typing import Final

ETS_6_SCHEMA_VERSION: Final = 21
ETS_5_7_SCHEMA_VERSION: Final = 20
ETS_5_6_SCHEMA_VERSION: Final = 14
ETS_4_2_SCHEMA_VERSION: Final = 11


MAIN_DPT: Final = "DPT"
MAIN_AND_SUB_DPT: Final = "DPST"

ETS4_PRODUCT_LANGUAGES: Final = [
    "cs-CZ",
    "da-DK",
    "de-DE",
    "el-GR",
    "en-US",
    "es-ES",
    "fi-FI",
    "fr-FR",
    "hu-HU",
    "is-IS",
    "it-IT",
    "ja-JP",
    "nb-NO",
    "nl-NL",
    "pl-PL",
    "pt-PT",
    "ro-RO",
    "ru-RU",
    "sk-SK",
    "sl-SI",
    "sv-SE",
    "tr-TR",
    "uk-UA",
    "zh-CN",
]
